#include "main.h"
#include "tsflow.h"

int g_initThreadCnt = 0;
bool g_startThreads = false;
int g_reqCnt = 0;
int g_tsReqCnt = 0;
int g_oscpReqCnt = 0;
unsigned param_threadNo = 1;
unsigned param_numOfQueries = 1;
int param_testmode = TM_TSA;
string param_tsa_authMode("basic");
string param_tsa_url;
string param_tsa_pfx;
string param_tsa_userName;
string param_tsa_passw;
string param_ocsp_authMode("basic");
string param_ocsp_url;
string param_ocsp_pfx;
string param_ocsp_userName;
string param_ocsp_passw;
string param_ocsp_cert;
string param_ocsp_issuer;
string param_qpattern("to");
int param_mask = 0;

void *myrealloc(void *ptr, size_t size)
{
  if(ptr)
    return realloc(ptr, size);
  else
    return malloc(size);
}

string my_printhex(const void* b, int size)
{
	char *buff = (char*)b;
	string res;
	char sum[128] = "";
	for (int l = 0; l < ((size/16)+1); l++) {
		char line[128] = "";
		char line2[32] = "";
		memset(line, 0, sizeof(line));
		memset(line2, 0, sizeof(line2));
		int linelen = 0;
		SNPRINTF(line, sizeof(line), "%04x: ", l*16);
		for (int c = 0; c < 16; c++) {
			char sch[4] = "   ";
			char sch2[2] = " ";
			if (((l*16) + c) < size) {
				char ch = buff[(l*16) + c];
				SNPRINTF(sch, 4, "%02x ", (0xFF & ch));
				if (int(ch) > 32) {
					sch2[0] = ch;
				} else {
					sch2[0] = '.';
				}
				linelen++;
			} else {
				break;
			}
			strcat(line, sch);			
			//if (linelen == 8) {
			//	strcat(line, "| ");
			//}
			strcat(line2, sch2);
		}
		if (linelen) {
			res.append(line);
			res.append(" [");
			res.append(line2);
			res.append("]\n");
		}
	}
	SNPRINTF(sum, sizeof(sum), " size: %d\n", size);
	res.append(sum);

	return res;
}

void OpenSSLDebug()
{
	int flags, line;
	const char *data, *file;
	unsigned long code;
	char str[256] = "";	

	if (param_mask & PARAM_VERBOSE) {
		code = ERR_get_error_line_data(&file, &line, &data, &flags);
		while (code) {
			ERR_error_string(code, str);
			printf("[OpenSSL] %s (%s:%s)\n", str, file, line);
			if (data && (flags & ERR_TXT_STRING)) {
				printf("[OpenSSL] error data: %s\n", data);
			}
			code = ERR_get_error_line_data(&file, &line, &data, &flags);
		}
		ERR_clear_error();
	}
}

int GetCertAndPrivKey(const string& path, const string& password // in
				, X509*& pCert, EVP_PKEY*& pPrivateKey) // out
{
	printf("GetCertAndPrivKey\n");
	X509* pClientCert = NULL;
	EVP_PKEY* pClientPrivateKey = NULL;

	BIO* pKeyBIO = BIO_new(BIO_s_file());
	if (pKeyBIO == NULL) {
		printf("BIO_new error\n");
		return 0;
	}

	FILE* pF = ::fopen(path.c_str(), "rb");
	if (!pF) {
		BIO_free_all(pKeyBIO);
		printf("could not read path: %s\n", path.c_str());
		return 0;
	}

	if (BIO_set_fp(pKeyBIO, pF, BIO_CLOSE) <= 0) {
		::fclose(pF);
		BIO_free_all(pKeyBIO);
		printf("could not read path: %s\n", path.c_str());
		return 0;
	}

	pClientPrivateKey = PEM_read_bio_PrivateKey(pKeyBIO,
		NULL, 0, (char*)password.c_str());

	if (pClientPrivateKey) {
		pClientCert = PEM_read_bio_X509_AUX(pKeyBIO, NULL, 0, NULL);
		if (pClientCert == NULL)
		{
			BIO_free_all(pKeyBIO);
			EVP_PKEY_free(pClientPrivateKey);
			::fclose(pF);
			printf("PEM_read_bio_X509_AUX error\n");
			return 0;
		}
	} else {
		// try PFX
		BIO_reset(pKeyBIO);
		PKCS12* p12 = d2i_PKCS12_bio(pKeyBIO, NULL);
		if (!p12) {
			BIO_free_all(pKeyBIO);
			::fclose(pF);
			printf("Read PFX error!\n");
			return 0;
		}
		printf("PKCS12_parse>\n");
		if (!PKCS12_parse(p12, password.c_str(), &pClientPrivateKey, &pClientCert, NULL)){
			PKCS12_free(p12);
			BIO_free_all(pKeyBIO);
			::fclose(pF);
			printf("PKCS12_parse error!\n");
			return 0;
		}
		printf("<PKCS12_parse\n");
		PKCS12_free(p12);
	} 
	BIO_free_all(pKeyBIO);
	::fclose(pF);

	pCert = pClientCert;
	pPrivateKey = pClientPrivateKey;

	return 1;
}

int get_params(int argc, char* argv[], string& msg)
{
	int argvcnt = 1;
	while (argvcnt < argc) {
		if (!strcmp("-m", argv[argvcnt])) {
			++argvcnt;
			if (!strcmp("tsa", argv[argvcnt])) {
				param_testmode = TM_TSA;
			} else if (!strcmp("ocsp", argv[argvcnt])) {
				param_testmode = TM_OCSP;
			} else if (!strcmp("mixed", argv[argvcnt])) {
				param_testmode = (TM_TSA | TM_OCSP);
			} else {
				msg = "test mode ?\n";
				return 0;
			}
		} else if (!strcmp("-t", argv[argvcnt])) {
			param_threadNo = atoi(argv[++argvcnt]);
		} else if (!strcmp("-n", argv[argvcnt])) {
			param_numOfQueries = atoi(argv[++argvcnt]);

		} else if (!strcmp("-a", argv[argvcnt])) {
			param_tsa_authMode = argv[++argvcnt];
		} else if (!strcmp("-h", argv[argvcnt])) {
			param_tsa_url = argv[++argvcnt];
		} else if (!strcmp("-u", argv[argvcnt])) {
			param_tsa_userName = argv[++argvcnt];
		} else if (!strcmp("-c", argv[argvcnt])) {
			param_tsa_pfx = argv[++argvcnt];
		} else if (!strcmp("-p", argv[argvcnt])) {
			param_tsa_passw = argv[++argvcnt];
		} else if (!strcmp("-r", argv[argvcnt])) {
			++argvcnt;
			if (!strcmp("sha256", argv[argvcnt])) {
				param_mask |= PARAM_TSAREQ_SHA256;
			} else if (!strcmp("sha1", argv[argvcnt])) {
				param_mask &= ~PARAM_TSAREQ_SHA256;
			} else {
				msg = "hash type of TS request ?\n";
				return 0;
			}

		} else if (!strcmp("-oa", argv[argvcnt])) {
			param_ocsp_authMode = argv[++argvcnt];
		} else if (!strcmp("-oh", argv[argvcnt])) {
			param_ocsp_url = argv[++argvcnt];
		} else if (!strcmp("-ou", argv[argvcnt])) {
			param_ocsp_userName = argv[++argvcnt];
		} else if (!strcmp("-oc", argv[argvcnt])) {
			param_ocsp_pfx = argv[++argvcnt];
		} else if (!strcmp("-op", argv[argvcnt])) {
			param_ocsp_passw = argv[++argvcnt];
		} else if (!strcmp("-cert", argv[argvcnt])) {
			param_ocsp_cert = argv[++argvcnt];
		} else if (!strcmp("-issuer", argv[argvcnt])) {
			param_ocsp_issuer = argv[++argvcnt];
		} else if (!strcmp("-or", argv[argvcnt])) {
			++argvcnt;
			if (!strcmp("sha256", argv[argvcnt])) {
				param_mask |= PARAM_OCSPREQ_SHA256;
			} else if (!strcmp("sha1", argv[argvcnt])) {
				param_mask &= ~PARAM_OCSPREQ_SHA256;
			} else {
				msg = "hash type of OCSP request ?\n";
				return 0;
			}

		} else if (!strcmp("-v", argv[argvcnt])) {			
			param_mask |= PARAM_VERBOSE;
		} else if (!strcmp("-s", argv[argvcnt])) {
			param_mask |= PARAM_VALIDATE;
		} else if (!strcmp("-qp", argv[argvcnt])) {
			param_qpattern = argv[++argvcnt];
		} else {
			return 0;
		}
		argvcnt++;
	}

	if (!param_threadNo) {
		msg = "number of threads?\n";
		return 0;
	}

	if (!param_numOfQueries) {
		msg = "number of queries?\n";
		return 0;
	}

	if (param_testmode & TM_TSA) {
		if (param_tsa_url.empty()) {
			msg = "TSA url?\n";
			return 0;
		}
		if (param_tsa_authMode == "basic") {
			if (param_tsa_userName.empty() || param_tsa_passw.empty()) {
				msg = "TSA user or pwd?\n";
				return 0;
			}
		} else if (param_tsa_authMode == "cert") { 
			if (param_tsa_pfx.empty() || param_tsa_passw.empty()) {
				msg = "TSA pfx or pwd?\n";
				return 0;
			}
		} else if (param_tsa_authMode == "none") { 
			param_tsa_userName.clear();
			param_tsa_pfx.clear();
		} else {
			msg = "TSA user authentication mode?\n";
			return 0;
		}
	}

	if (param_testmode & TM_OCSP) {
		if (param_ocsp_url.empty()) {
			msg = "OCSP url?";
			return 0;
		}

		if (param_ocsp_cert.empty()) {
			msg = "OCSP cert?";
			return 0;
		}

		if (param_ocsp_issuer.empty()) {
			msg = "OCSP issuer cert?";
			return 0;
		}

		if (param_ocsp_authMode == "basic") {
			if (param_ocsp_userName.empty()) {
				if (param_tsa_userName.empty()) {
					msg = "OCSP user name?\n";
					return 0;
				} else {
					param_ocsp_userName = param_tsa_userName;
				}
			}			
		} else if (param_ocsp_authMode == "cert") {
			if (param_ocsp_pfx.empty()) {
				if (param_tsa_pfx.empty()) {
					msg = "OCSP pfx?\n";
					return 0;
				} else {
					param_ocsp_pfx = param_tsa_pfx;
				}
			}			
		} else if (param_ocsp_authMode == "none") {
			param_ocsp_userName.clear();
			param_ocsp_pfx.clear();
		} else {
			msg = "OCSP user authentication mode?\n";
		}

		if (param_ocsp_passw.empty()) {
			if (param_tsa_passw.empty()) {
				msg = "OCSP user pwd?\n";
				return 0;
			} else {
				param_ocsp_passw = param_tsa_passw;
			}
		}
	}
	
	if (param_testmode == (TM_TSA | TM_OCSP)) {
		if (param_qpattern.empty()) {
			msg = "query pattern?\n";
			return 0;
		}
		for (int i = 0; i < param_qpattern.length(); i++) {
			if ((param_qpattern[i] != 't') &&
				(param_qpattern[i] != 'o')) {
				msg = "invalid char in query pattern\n";
				return 0;
			}
		}
	}

	return 1;
}

int main(int argc, char* argv[])
{	
	vector<CTest*> vTests;

	OpenSSL_add_all_algorithms();

	printf("TSFlow - The TSA/OCSP responder stress tool, version 1.0 (build %d)\n\n", __COUNTER__);

	// get params --------

	string msg;
	if (!get_params(argc, argv, msg)) {
		printf("Usage: tsflow [-m TESTMODE] [-n NUMOFTS] [-h url] [-t THREADNO] [-a|oa AUTHMODE] [-c|oc file] [-u|ou USERNAME] [-p|op PASSW] [-v] [-s] [-r] [-cert file] [-issuer file] [-qp PATTERN]\n");
		printf("\n");
		printf("  -m        testing mode (def:TSA)\n");
		printf("               \"tsa\"     -tsa server test only(default)\n");
		printf("               \"ocsp\"    -ocsp server test only\n");
		printf("               \"mixed\"   -mixed mode, both tsa and ocsp tests will be\n");
		printf("  -n        number of queries (def:1)\n");
		printf("  -t        number of test threads (def:1)\n");
		printf("  -h        TimeStamp server\n");
		printf("  -a        type of authentication: (def:\"basic\")\n");
		printf("               \"basic\"   -basic authentication, USERNAME/PASSW\n");
		printf("               \"cert\"    -certificate base authentication, PFXPATH/PASSW\n");
		printf("               \"none\"\n");
		printf("  -u        user name\n");
		printf("  -c        path of cert file\n");
		printf("  -p        password of user or cert file\n");
		printf("  -v        verbose (def:no)\n");
		printf("  -s        validate signature (def:no, asn1 parse only)\n");
		printf("  -r        TS request type:\n");
		printf("               \"sha1\"    (default)\n");
		printf("               \"sha256\"\n");
		printf("  -qp       query pattern: Simple string which is contains the pattern of TS [t] and OCSP [o] queries in order.\n");
		printf("              (default:\"to\")\n");
		printf("\n");
		printf(" OCSP server connection parameters for user authentication: \n");
		printf("\n");
		printf("  -oh       see above: [-h]\n");
		printf("  -oa       see above: [-a] (def: inherited by -a)\n");
		printf("  -ou       see above: [-u] (def: inherited by -u)\n");
		printf("  -oc       see above: [-c] (def: inherited by -c)\n");
		printf("  -op       see above: [-p] (def: inherited by -p)\n");
		printf("  -or       see above: [-r] (def: \"sha1\")\n");
		printf("  -cert     certificate to check (.cer)\n");
		printf("  -issuer   issuer certificate (.cer)\n");
		printf("\n");

		if (!msg.empty()) {
			printf("\n  %s", msg.c_str());
		}
		return -1;
	}

	printf("--------------------------------------------------------------\n");
	printf("  number of queries:%d\n", param_numOfQueries);
	printf("  threadNo:         %d\n", param_threadNo);
	if (param_mask & PARAM_VERBOSE) {
		printf("  verbose:          yes\n");
	}
	if (param_mask & PARAM_VALIDATE) {
		printf("  validate:         yes\n");
	}
	if (param_testmode & TM_TSA) {
		printf("\n");
		printf("  TSA test params:\n");
		printf("  ----------------\n");
		printf("    URL:              %s\n", param_tsa_url.c_str());
		printf("    authMode:         %s\n", param_tsa_authMode.c_str());
		printf("    request type:     %s\n", (param_mask & PARAM_TSAREQ_SHA256) ? "sha256" : "sha1");
	}
	if (param_testmode & TM_OCSP) {
		printf("\n");
		printf("  OCSP test params:\n");
		printf("  -----------------\n");
		printf("    URL:              %s\n", param_ocsp_url.c_str());
		printf("    authMode:         %s\n", param_ocsp_authMode.c_str());
		printf("    request type:     %s\n", (param_mask & PARAM_OCSPREQ_SHA256) ? "sha256" : "sha1");
	}
	if (param_testmode == (TM_TSA | TM_OCSP)) {
		printf("\n");
		printf("  Mixed-mode test params:\n");
		printf("  -----------------------\n");
		printf("    query pattern:    [%s]\n", param_qpattern.c_str());
	}
	printf("\n");

// ---------

	g_initThreadCnt = param_threadNo;
	g_reqCnt = param_numOfQueries;
	g_tsReqCnt = param_numOfQueries;
	g_oscpReqCnt = param_numOfQueries;

	if (param_testmode & TM_OCSP) {
		if (!COCSPTest::createRequest()) {
			printf("Error occured while creating ocsp request. Please check [-cert] and [-issuer] parameters.");
			return -1;
		}
	}

	for (unsigned tn = 0; tn < param_threadNo; tn++) {
		CTest* pT = new CTest(param_threadNo, tn);
		if (pT) {
			pT->startProc();
			vTests.push_back(pT);
		} else {
			printf("oops!");
			return -1;
		}
	}
	
	// waiting for init
	while (g_initThreadCnt > 0) {
		Sleep(50);
	}

// >>> ts_first
time_t ts_first = time(0);
printf(">>> start:%lld\n", ts_first);

	// run
	g_startThreads = true;

	while (1) {
		unsigned job = 0;
#if defined(WIN32)	
		for (vector<CTest*>::iterator pT = vTests.begin(); pT != vTests.end(); pT++) {
			job += (*pT)->waitForProc(1 + (1000 / param_threadNo));
		}
		if (/*is there any*/job) {
			printf(".");
		} else {
			break;
		}
#else 
		for (vector<CTest*>::iterator pT = vTests.begin(); pT != vTests.end(); pT++) {
			job += (*pT)->m_isRun;
		}		
		if (/*is there any*/job) {
			if (t != time(0)) {
			    t = time(0);
    			    putc('.', stdout);
			    fflush(stdout);
			}
		} else {
			break;
		}
#endif		
	}
	printf("\nDone.\n");

// >>> ts_last
time_t ts_last = time(0);
printf(">>> end:%lld\n", ts_last);

	printf("\nsave files ..");
	unsigned allTryTS = 0;
	unsigned allTryOCSP = 0;
	unsigned allGoodTS = 0;
	unsigned allGoodOCSP = 0;
	for (vector<CTest*>::iterator pT = vTests.begin(); pT != vTests.end(); pT++) {
		if (param_testmode & TM_TSA) {
			(*pT)->CTSATest::saveFiles();
		}
		if (param_testmode & TM_OCSP) {
			(*pT)->COCSPTest::saveFiles();
		}
		allTryTS += (*pT)->m_allTryTS;
		allTryOCSP += (*pT)->m_allTry;
		allGoodTS += (*pT)->m_allGoodTS;
		allGoodOCSP += (*pT)->m_allGood;
		printf(".");
	}	
	printf(" done\n");

	printf("--------------------------------------------------------------\n");
	printf("start:%lld ", ts_first);
	printf("end:%lld  ---  ", ts_last);
	printf("diff:%lld sec\n", ts_last - ts_first);
	if (param_testmode & TM_TSA) {
		 printf("  TSA  responses: valid=%d try=%d - %.1f%%\n", allGoodTS, allTryTS, ((100.0*allGoodTS)/allTryTS));
	}
	if (param_testmode & TM_OCSP) {
		 printf("  OCSP responses: valid=%d try=%d - %.1f%%\n", allGoodOCSP, allTryOCSP, ((100.0*allGoodOCSP)/allTryOCSP));
	}

	return 0;
}
