#include <fstream>

using namespace std;

#include "ocsptest.h"

const char CT_OCSPQ[] = "Content-Type: application/ocsp-request";

COCSPTest::COCSPTest(int id)
:m_id(id)
,m_curl(NULL)
,m_headers(NULL)
,m_allGood(0)
,m_allTry(0)
{
	memset(&m_SSLCtx, 0, sizeof(m_SSLCtx));
}


MemoryStruct COCSPTest::m_request(NULL, 0);

COCSPTest::~COCSPTest()
{
}

int COCSPTest::curl_debug_clbk(CURL *curl, curl_infotype infotype, char *data, size_t data_len, void *x)
{
	switch (infotype) {
	case CURLINFO_TEXT:	
		if (data_len) {
			string s(data, data_len);
			char ch = s[s.length() - 1];
			while ((ch == '\r') 
				|| (ch == '\n')) {
				s = s.substr(0, s.length() - 1);
				ch = s[s.length() - 1];
			}
			printf("[OCSP] curl: %s\n", s.c_str());
		} 
		break;
	//case CURLINFO_HEADER_IN:
	//	printf("curl: CURLINFO_HEADER_IN\n");
	//	printf(my_printhex(data, data_len).c_str());
	//	break;
	//case CURLINFO_HEADER_OUT:
	//	printf("curl: CURLINFO_HEADER_OUT\n");
	//	printf(my_printhex(data, data_len).c_str());
	//	break;
	//case CURLINFO_DATA_IN:
	//	printf("curl: CURLINFO_DATA_IN\n");
	//	printf(my_printhex(data, data_len).c_str());
	//	break;
	//case CURLINFO_DATA_OUT:
	//	printf("curl: CURLINFO_DATA_OUT\n");
	//	printf(my_printhex(data, data_len).c_str());
	//	break;
	}

	return 0; // This function must return 0.
} 

// curl answer writing callback
size_t COCSPTest::WriteMemoryCallback(void *ptr, size_t size, size_t nmemb, void *data)
{
	size_t realsize = size * nmemb;
	struct MemoryStruct *mem = (struct MemoryStruct *)data;

	mem->memory = (unsigned char *)myrealloc(mem->memory, mem->size + realsize + 1);
	if (mem->memory) {
		memcpy(&(mem->memory[mem->size]), ptr, realsize);
		mem->size += realsize;
		mem->memory[mem->size] = 0;
	}
	return realsize;
} 

CURLcode COCSPTest::SSLCtxCallback(CURL * curl, void * sslctx, void * parm)
{
	SSLCTXPARAM * p = (SSLCTXPARAM *) parm;
	SSL_CTX * ctx = (SSL_CTX *) sslctx ;

	// vigyazz! ha itt ujabb felttelt adsz, nzd meg azt az esetet is amikor 
	// nem kell meghvdnia ennek a callback-nek de mgis megtrtnik. (pl.: curl_easy_duphandle)
	// nzd meg ezt: /*TAG:SSLCtxCallback_param*/ 
	if (p->usercert != NULL && p->pkey != NULL) {
		if (SSL_CTX_use_certificate(ctx, p->usercert) <= 0) {
			printf("[OCSP] SSL context callback error: XSL_SSL_CTX_USE_CERT_ERR.\n");
			throw;
		}
		if (SSL_CTX_use_PrivateKey(ctx, p->pkey) <= 0) {
			printf("[OCSP] SSL context callback error: XSL_SSL_CTX_USE_KEY_ERR.\n");
			throw;
		}
		if (!SSL_CTX_check_private_key(ctx)) {
			printf("[OCSP] SSL context callback error: XSL_SSL_CTX_CERT_AND_KEY_MISMATCH.\n");
			throw;
		}
	}
	//SSL_CTX_set_verify(ctx ,SSL_VERIFY_NONE,verify_callback);
	return CURLE_OK;  
}

int COCSPTest::initConnection()
{
	m_curl = curl_easy_init();

	curl_easy_setopt(m_curl, CURLOPT_NOSIGNAL, 1);

	if (param_mask & PARAM_VERBOSE) {
		curl_easy_setopt(m_curl, CURLOPT_VERBOSE, 1);
		curl_easy_setopt(m_curl, CURLOPT_DEBUGFUNCTION, curl_debug_clbk);
	}
	//curl_easy_setopt(m_curl, CURLOPT_SOCKOPTFUNCTION, curl_sockopt_clbk);
	//curl_easy_setopt(m_curl, CURLOPT_HEADER, 1);

	//long httpauth_avail = CURLAUTH_ANY;
	//curl_easy_getinfo(m_curl, CURLINFO_HTTPAUTH_AVAIL, &httpauth_avail);
	//printf("http_authvail: %d\n", httpauth_avail);
	//curl_easy_setopt(m_curl, CURLOPT_HTTPAUTH, httpauth_avail); 
	X509* pClientCert = NULL;
	EVP_PKEY* pClientPrivateKey = NULL;

	if (!param_ocsp_userName.empty()) {
		printf("[OCSP] SSL basic authentication.\n");
		m_userPass = param_ocsp_userName + ":" + param_ocsp_passw;
		curl_easy_setopt(m_curl, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
		curl_easy_setopt(m_curl, CURLOPT_USERPWD, m_userPass.c_str());
	} else if (!param_ocsp_pfx.empty()) {
		printf("[OCSP] SSL cert authentication.\n");

		if (!GetCertAndPrivKey(param_ocsp_pfx, param_ocsp_passw, pClientCert, pClientPrivateKey)) {
			return 0;
		}

		m_SSLCtx.verbose = 0;
		m_SSLCtx.curl = m_curl;
		m_SSLCtx.p12file = NULL;
		m_SSLCtx.pst = NULL;
		m_SSLCtx.pkey = pClientPrivateKey;
		m_SSLCtx.usercert = pClientCert;
		m_SSLCtx.ca = NULL;

		curl_easy_setopt(m_curl, CURLOPT_SSL_CTX_FUNCTION, SSLCtxCallback);
		curl_easy_setopt(m_curl, CURLOPT_SSL_CTX_DATA, &m_SSLCtx);
		//curl_easy_setopt(m_curl, CURLOPT_HTTPAUTH, CURLAUTH_ANY);
		curl_easy_setopt(m_curl, CURLOPT_HTTPAUTH, CURLAUTH_ANYSAFE);
	}

	curl_easy_setopt(m_curl, CURLOPT_SSL_VERIFYPEER, 0);

	//Set sContentType
	m_sContentType = CT_OCSPQ;
	m_headers = NULL;
	if (m_sContentType.length()) {
		m_headers = curl_slist_append(m_headers, m_sContentType.c_str());
	}
	if (m_headers) {
		curl_easy_setopt(m_curl, CURLOPT_HTTPHEADER, m_headers); 
	}

	m_sURL = param_ocsp_url;
	curl_easy_setopt(m_curl, CURLOPT_WRITEFUNCTION, WriteMemoryCallback);
	curl_easy_setopt(m_curl, CURLOPT_USERAGENT, "tsflow/1.0"); 
	curl_easy_setopt(m_curl, CURLOPT_URL, m_sURL.c_str());

	curl_easy_setopt(m_curl, CURLOPT_POSTFIELDS, m_request.memory);
	curl_easy_setopt(m_curl, CURLOPT_POSTFIELDSIZE, m_request.size);

	return 1;
}

X509* getCert(string fname) 
{
	X509* cert = NULL;
	BIO *pKeyBIO = NULL;

	pKeyBIO = BIO_new(BIO_s_file());
	if ((pKeyBIO == NULL) || (BIO_read_filename(pKeyBIO, fname.c_str()) <= 0)) {
		return 0;
	}
	
	cert = d2i_X509_bio(pKeyBIO, NULL);
	if (cert == NULL) {
		// try PFX
		BIO_read_filename(pKeyBIO, fname.c_str());
		cert = PEM_read_bio_X509_AUX(pKeyBIO, NULL, 0, NULL);		
	}

	BIO_free(pKeyBIO);
	return cert;
}	

int COCSPTest::createRequest()
{
	int res = 0;
	string sHashAlgorithm = "sha1";
	X509* cert = NULL;
	X509* issuer = NULL;
	const EVP_MD *dgst = NULL;
	OCSP_CERTID *pCertId = NULL;
	OCSP_ONEREQ *pOneReq = NULL;
	OCSP_REQUEST *pReq = NULL;

	X509* pClientCert = NULL;
	EVP_PKEY* pClientPrivateKey = NULL;

	if (param_mask & PARAM_OCSPREQ_SHA256) {
		sHashAlgorithm = "sha256";
	}

	BIO* pKeyBIO = BIO_new(BIO_s_file());
	if (pKeyBIO == NULL) {
		printf("BIO_new error\n");
		return 0;
	}

	cert = getCert(param_ocsp_cert);
	if (cert) {
		issuer = getCert(param_ocsp_issuer);
		if (issuer) {
			pReq = OCSP_REQUEST_new();
			if (pReq) {
				dgst = EVP_get_digestbyname(sHashAlgorithm.c_str());
				if (dgst) {	
					pCertId = OCSP_cert_to_id(dgst, cert, issuer);
					if (pCertId) {
						pOneReq = OCSP_request_add0_id(pReq, pCertId);
						if (pOneReq) {							
							m_request.size = i2d_OCSP_REQUEST(pReq, NULL);
							if (m_request.size) {
								unsigned char *pDummy = m_request.memory = new unsigned char[m_request.size];
								if (m_request.memory) {
									if (i2d_OCSP_REQUEST(pReq, &pDummy) == m_request.size) {		
										//OK
										fstream outf("ocsp_req.p7s", ios::out | ios::binary);
										outf.write((char*)m_request.memory, (std::streamsize)m_request.size);

										res = 1;
									}
								} else {
									m_request.size = 0;
								}
							}
						}
					}
				}
				OCSP_REQUEST_free(pReq);
			}
			X509_free(issuer);
		}
		X509_free(cert);
	}
	return res;
}

int COCSPTest::getOne()
{
	m_allTry++;

	MemoryStruct* pMS = new MemoryStruct(NULL,0);
	if (pMS) {
		m_pmBuffs.push_back(pMS);
		curl_easy_setopt(m_curl, CURLOPT_WRITEDATA, (void *)pMS);

		curl_easy_perform(m_curl);
		//printf(my_printhex(mBuff.memory, mBuff.size).c_str());		
	}
	return 1;
}

void COCSPTest::saveFiles()
{
	unsigned fnid = 0;
	vector<MemoryStruct*>::iterator ipBuff = m_pmBuffs.begin();
	for (; ipBuff != m_pmBuffs.end(); ipBuff++) {
		char fname[64] = "";
		_snprintf(fname, 64, "ocsp_%02d_%04d.p7s", m_id, fnid++);

		{
			OCSP_RESPONSE* pDummy = NULL;
			OCSP_RESPONSE* pResponse = NULL;
			const unsigned char* pTmp = (const unsigned char*)(*ipBuff)->memory;
			long lDataLength = (long)(*ipBuff)->size;

			pResponse = d2i_OCSP_RESPONSE(&pDummy, &pTmp, lDataLength);				
			if (pResponse) {
				int res = OCSP_response_status(pResponse);
				if (0 == res) {
					m_allGood++;
				} else {
					if (param_mask & PARAM_VERBOSE) {
						printf("[OCSP] ERROR: [%s] - invalid response status\n", fname);
					}
				}
				OCSP_RESPONSE_free(pResponse);
			}
		}

		fstream outf(fname, ios::out | ios::binary);
		outf.write((char*)(*ipBuff)->memory, (std::streamsize)(*ipBuff)->size);
	}
}
