#include <fstream>

using namespace std;

#include "tsatest.h"
#include <openssl/applink.c>

//#define BUFFINITSIZE 2800

const char TSREQ[] = "\x30\x34\x02\x01\x01\x30\x21\x30\x09\x06\x05\x2b\x0e\x03\x02\x1a\x05\x00\x04\x14\x6b\x64\x73\xa9\xff\x13\xf2\x41\x92\x10\xfb\x24\x2a\xec\xa7\xcf\x30\xba\xab\x07\x02\x09\x00\xfd\x26\x6c\x94\xe1\x1a\x27\x3a\x01\x01\xff";
const char TSREQ_SHA256[] = "\x30\x43\x02\x01\x01\x30\x31\x30\x0d\x06\x09\x60\x86\x48\x01\x65\x03\x04\x02\x01\x05\x00\x04\x20\x18\x90\xd9\x75\xcf\xe8\xab\xbb\x98\x50\x2d\x4d\x5c\xcc\x59\x23\xee\x8e\xba\x27\xde\x70\x33\x04\xc5\xc9\x43\xfc\x99\xa6\x17\x6f\x02\x08\x4f\x35\x5a\x69\xec\x01\x3f\x01\x01\x01\xff";
const char CT_TSAQ[] = "Content-Type: application/timestamp-query";

CTSATest::CTSATest(int id)
	:m_id(id)
	,m_curl(NULL)
	,m_headers(NULL)
	,m_allGoodTS(0)
	,m_allTryTS(0)
{
	memset(&m_SSLCtx, 0, sizeof(m_SSLCtx));
}

CTSATest::~CTSATest()
{
	if (m_headers) {
		curl_slist_free_all(m_headers);
	}
	if (m_curl) {
		curl_easy_cleanup(m_curl);
	}

	vector<MemoryStruct*>::iterator ipBuff = m_pmBuffs.begin();
	for (; ipBuff != m_pmBuffs.end(); ipBuff++) {
		delete[] (*ipBuff);
	}
	m_pmBuffs.clear();
}

int CTSATest::curl_debug_clbk(CURL *curl, curl_infotype infotype, char *data, size_t data_len, void *x)
{
	switch (infotype) {
	case CURLINFO_TEXT:	
		if (data_len) {
			string s(data, data_len);
			char ch = s[s.length() - 1];
			while ((ch == '\r') 
				|| (ch == '\n')) {
				s = s.substr(0, s.length() - 1);
				ch = s[s.length() - 1];
			}
			printf("[TS]   curl: %s\n", s.c_str());
		} 
		break;
	//case CURLINFO_HEADER_IN:
	//	printf("curl: CURLINFO_HEADER_IN\n");
	//	printf(my_printhex(data, data_len).c_str());
	//	break;
	//case CURLINFO_HEADER_OUT:
	//	printf("curl: CURLINFO_HEADER_OUT\n");
	//	printf(my_printhex(data, data_len).c_str());
	//	break;
	//case CURLINFO_DATA_IN:
	//	printf("curl: CURLINFO_DATA_IN\n");
	//	printf(my_printhex(data, data_len).c_str());
	//	break;
	//case CURLINFO_DATA_OUT:
	//	printf("curl: CURLINFO_DATA_OUT\n");
	//	printf(my_printhex(data, data_len).c_str());
	//	break;
	}

	return 0; // This function must return 0.
} 

// curl answer writing callback
size_t CTSATest::WriteMemoryCallback(void *ptr, size_t size, size_t nmemb, void *data)
{
	size_t realsize = size * nmemb;
	struct MemoryStruct *mem = (struct MemoryStruct *)data;

	mem->memory = (unsigned char *)myrealloc(mem->memory, mem->size + realsize + 1);
	if (mem->memory) {
		memcpy(&(mem->memory[mem->size]), ptr, realsize);
		mem->size += realsize;
		mem->memory[mem->size] = 0;
	}
	return realsize;
} 

CURLcode CTSATest::SSLCtxCallback(CURL * curl, void * sslctx, void * parm)
{
	SSLCTXPARAM * p = (SSLCTXPARAM *) parm;
	SSL_CTX * ctx = (SSL_CTX *) sslctx ;

	// vigyazz! ha itt ujabb felttelt adsz, nzd meg azt az esetet is amikor 
	// nem kell meghvdnia ennek a callback-nek de mgis megtrtnik. (pl.: curl_easy_duphandle)
	// nzd meg ezt: /*TAG:SSLCtxCallback_param*/ 
	if (p->usercert != NULL && p->pkey != NULL) {
		if (SSL_CTX_use_certificate(ctx, p->usercert) <= 0) {
			printf("[TS]   SSL context callback error: XSL_SSL_CTX_USE_CERT_ERR.\n");
			throw;
		}
		if (SSL_CTX_use_PrivateKey(ctx, p->pkey) <= 0) {
			printf("[TS]   SSL context callback error: XSL_SSL_CTX_USE_KEY_ERR.\n");
			throw;
		}
		if (!SSL_CTX_check_private_key(ctx)) {
			printf("[TS]   SSL context callback error: XSL_SSL_CTX_CERT_AND_KEY_MISMATCH.\n");
			throw;
		}
	}
	//SSL_CTX_set_verify(ctx ,SSL_VERIFY_NONE,verify_callback);
	return CURLE_OK;  
}

int CTSATest::initConnection()
{
	m_curl = curl_easy_init();

	curl_easy_setopt(m_curl, CURLOPT_NOSIGNAL, 1);

	if (param_mask & PARAM_VERBOSE) {
		curl_easy_setopt(m_curl, CURLOPT_VERBOSE, 1);
		curl_easy_setopt(m_curl, CURLOPT_DEBUGFUNCTION, curl_debug_clbk);
	}
	//curl_easy_setopt(m_curl, CURLOPT_SOCKOPTFUNCTION, curl_sockopt_clbk);
	//curl_easy_setopt(m_curl, CURLOPT_HEADER, 1);

	//long httpauth_avail = CURLAUTH_ANY;
	//curl_easy_getinfo(m_curl, CURLINFO_HTTPAUTH_AVAIL, &httpauth_avail);
	//printf("http_authvail: %d\n", httpauth_avail);
	//curl_easy_setopt(m_curl, CURLOPT_HTTPAUTH, httpauth_avail); 
	X509* pClientCert = NULL;
	EVP_PKEY* pClientPrivateKey = NULL;

	if (!param_tsa_userName.empty()) {
		printf("[TS]   SSL basic authentication.\n");
		m_userPass = param_tsa_userName + ":" + param_tsa_passw;
		curl_easy_setopt(m_curl, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
		curl_easy_setopt(m_curl, CURLOPT_USERPWD, m_userPass.c_str());
	} else if (!param_tsa_pfx.empty()) {
		printf("[TS]   SSL cert authentication.\n");

		if (!GetCertAndPrivKey(param_tsa_pfx, param_tsa_passw, pClientCert, pClientPrivateKey)) {
			return 0;
		}

		m_SSLCtx.verbose = 0;
		m_SSLCtx.curl = m_curl;
		m_SSLCtx.p12file = NULL;
		m_SSLCtx.pst = NULL;
		m_SSLCtx.pkey = pClientPrivateKey;
		m_SSLCtx.usercert = pClientCert;
		m_SSLCtx.ca = NULL;

		curl_easy_setopt(m_curl, CURLOPT_SSL_CTX_FUNCTION, SSLCtxCallback);
		curl_easy_setopt(m_curl, CURLOPT_SSL_CTX_DATA, &m_SSLCtx);
		//curl_easy_setopt(m_curl, CURLOPT_HTTPAUTH, CURLAUTH_ANY);
		curl_easy_setopt(m_curl, CURLOPT_HTTPAUTH, CURLAUTH_ANYSAFE);
	}

	curl_easy_setopt(m_curl, CURLOPT_SSL_VERIFYPEER, 0);

	//Set sContentType
	m_sContentType = CT_TSAQ;
	m_headers = NULL;
	if (m_sContentType.length()) {
		m_headers = curl_slist_append(m_headers, m_sContentType.c_str());
	}
	if (m_headers) {
		curl_easy_setopt(m_curl, CURLOPT_HTTPHEADER, m_headers); 
	}

	m_sURL = param_tsa_url;
	curl_easy_setopt(m_curl, CURLOPT_WRITEFUNCTION, WriteMemoryCallback);
	curl_easy_setopt(m_curl, CURLOPT_USERAGENT, "tsflow/1.0"); 
	curl_easy_setopt(m_curl, CURLOPT_URL, m_sURL.c_str());

	curl_easy_setopt(m_curl, CURLOPT_POST, 1); 
	if (param_mask & PARAM_TSAREQ_SHA256) {
		curl_easy_setopt(m_curl, CURLOPT_POSTFIELDS, TSREQ_SHA256);
		curl_easy_setopt(m_curl, CURLOPT_POSTFIELDSIZE, sizeof(TSREQ_SHA256)-1/*NULL terminated*/);
	} else {
		curl_easy_setopt(m_curl, CURLOPT_POSTFIELDS, TSREQ);
		curl_easy_setopt(m_curl, CURLOPT_POSTFIELDSIZE, sizeof(TSREQ)-1/*NULL terminated*/);
	}

	return 1;
}

int CTSATest::getOne()
{
	m_allTryTS++;

	MemoryStruct* pMS = new MemoryStruct(NULL,0);
	if (pMS) {
		m_pmBuffs.push_back(pMS);
		curl_easy_setopt(m_curl, CURLOPT_WRITEDATA, (void *)pMS);
		
		curl_easy_perform(m_curl);
		//printf(my_printhex(mBuff.memory, mBuff.size).c_str());		
	}
	return 1;
}

void CTSATest::saveFiles()
{
	unsigned fnid = 0;
	vector<MemoryStruct*>::iterator ipBuff = m_pmBuffs.begin();
	for (; ipBuff != m_pmBuffs.end(); ipBuff++) {

		char fname[64] = "";
		_snprintf(fname, 64, "ts_%02d_%04d.p7s", m_id, fnid++);

		{
			TS_RESP* pDummy = NULL;
			TS_RESP* pTSResponse = NULL;
			const unsigned char* pTmp = (const unsigned char*)(*ipBuff)->memory;
			long lDataLength = (long)(*ipBuff)->size;

			pTSResponse = d2i_TS_RESP(&pDummy, &pTmp, lDataLength);				
			if (pTSResponse) {
				PKCS7 * pPKCS7 = TS_RESP_get_token(pTSResponse);
				if (pPKCS7) {
					if (param_mask & PARAM_VALIDATE) {
						STACK_OF(PKCS7_SIGNER_INFO) *sinfos = PKCS7_get_signer_info(pPKCS7);
						if (sinfos && sk_PKCS7_SIGNER_INFO_num(sinfos)) {
							int good = 0;
							for (int i = 0; i < sk_PKCS7_SIGNER_INFO_num(sinfos); i++) {
								PKCS7_SIGNER_INFO *si = sk_PKCS7_SIGNER_INFO_value(sinfos, i);

								PKCS7_ISSUER_AND_SERIAL* pIAS = si->issuer_and_serial;
								X509* pSigner = pSigner = X509_find_by_issuer_and_serial(pPKCS7->d.sign->cert, pIAS->issuer, pIAS->serial);

								if (pSigner) {
									BIO	*p7bio = PKCS7_dataInit(pPKCS7, NULL);
									char buf[4096];
									while ((i = BIO_read(p7bio, buf, sizeof(buf))) > 0);

									int res = PKCS7_signatureVerify(p7bio, pPKCS7, si, pSigner);
									if (res > 0) {
										good++;
									}
									BIO_free_all(p7bio);
								} else {
									OpenSSLDebug();
									printf("[TS]   ERROR: [%s] - signer not found\n", fname);
								}
							}
							if (good) {
								m_allGoodTS++;
							} else {
								OpenSSLDebug();
								printf("[TS]   ERROR: [%s] - there is no any signature validated\n", fname);
							}
						} else {
							OpenSSLDebug();
							printf("[TS]   ERROR: [%s] - no PKCS7_SIGNER_INFO!\n", fname);
						}
					} else { // param_mask & PARAM_VALIDATE
						m_allGoodTS++;
					}
				} else {
					OpenSSLDebug();
					printf("[TS]   ERROR: [%s] - PKCS7 token not found!\n", fname);
				}
				TS_RESP_free(pTSResponse);
			} else {
				OpenSSLDebug();
				printf("[TS]   ERROR: [%s] - invalid response data!\n", fname);
			}
		}

		fstream outf(fname, ios::out | ios::binary);
		outf.write((char*)(*ipBuff)->memory, (std::streamsize)(*ipBuff)->size);
	}
}
