#include <stdio.h>
#include <string>
#include <fstream>
#include <vector>

#include "tsflow.h"

using namespace std;

// ---------------

CTest::CTest(int nThreads, int id)
	:CTSATest(id)
	,COCSPTest(id)
	,m_id(id)
	,m_nThreads(nThreads)
	,m_hThread(0)
	,m_threadID(0)
	,m_isRun(false)
{
	;
}

CTest::~CTest()
{
	if (m_hThread) {
#if defined(WIN32)	
		CloseHandle(m_hThread);
#else
		//
#endif		
	}
}

int CTest::run()
{
	if (param_testmode & TM_TSA) {
		if (!CTSATest::initConnection()) {
			return 0;
		}
	}

	if (param_testmode & TM_OCSP) {
		if (!COCSPTest::initConnection()) {
			return 0;
		}
	}

	// init done:
	g_initThreadCnt--;

	// waiting for init
	while (g_initThreadCnt > 0) {
		Sleep(50);
	}

	// waiting for start
	while (!g_startThreads) {
		Sleep(50);
	}

	printf("id#%d - run\n", m_id);

	if (param_testmode == (TM_TSA | TM_OCSP)) {
		while (g_reqCnt > 0) {
			g_reqCnt--;
			for (int i = 0; i < param_qpattern.length(); i++) {
				if (param_qpattern[i] == 't') {
					CTSATest::getOne();
				} else if (param_qpattern[i] == 'o') {
					COCSPTest::getOne();
				}
			}
		}
	} else if (param_testmode & TM_TSA) {
		while (g_tsReqCnt-- > 0) {
			CTSATest::getOne();
		}
	} else if (param_testmode & TM_OCSP) {
		while (g_oscpReqCnt-- > 0) {
			COCSPTest::getOne();
		}
	}
	return 0;
}

DWORD MYCALLTYPE CTest::threadFunct(LPVOID lpThreadParameter)
{
	CTest *pThis = (CTest*)lpThreadParameter;
	if (pThis) {
		printf("id#%d - init\n", pThis->m_id);
		pThis->m_isRun = true;
		pThis->run();		
		pThis->m_isRun = false;
	}
	return 1;
}

void CTest::startProc() 
{
#if defined(WIN32)    
	m_threadID = 0;
	m_hThread = (HANDLE)CreateThread(
		NULL, 
		0, 
		/*(LPTHREAD_START_ROUTINE)*/threadFunct, 
		(LPVOID)this, 
		0, 
		(LPDWORD)&(m_threadID));
#else 
	pthread_create(&m_hThread, NULL, (void*(*)(void*))&threadFunct, (void*)this);
#endif
}

unsigned CTest::waitForProc(unsigned ms) 
{
#if defined(WIN32)    
	return WAIT_TIMEOUT == WaitForSingleObject(m_hThread, ms);
#endif
}
